from random import Random

from sampo.generator.environment.contractor import get_contractor
from sampo.generator.pipeline.extension import extend_names, extend_resources
from sampo.generator.pipeline.project import get_small_graph, get_graph
from sampo.generator.types import SyntheticGraphType
from sampo.schemas.graph import WorkGraph


class SimpleSynthetic:
    def __init__(self, seed: int | None) -> None:
        self._rand = Random(seed)

    def small_work_graph(self, cluster_name: str | None = 'C1') -> WorkGraph:
        return get_small_graph(cluster_name, self._rand)

    def work_graph(self, mode: SyntheticGraphType | None = SyntheticGraphType.General,
                   bottom_border: int | None = 0,
                   top_border: int | None = 0) -> WorkGraph:
        return get_graph(mode=mode, bottom_border=bottom_border, top_border=top_border, rand=self._rand)

    def contactor(self, pack_worker_count: float):
        return get_contractor(pack_worker_count, rand=self._rand)

    def advanced_work_graph(self, works_count_top_border: int, uniq_works: int, uniq_resources: int, scaler: int = 5):
        wg = self.work_graph(top_border=works_count_top_border)
        wg = extend_names(uniq_works, wg, self._rand)
        wg = extend_resources(uniq_resources, wg, self._rand)
        return wg



from collections import defaultdict
from dataclasses import dataclass, field
from functools import partial
from typing import Tuple, Optional, List, Union, Dict, Callable

import numpy as np
from numpy import ndarray
import scipy.optimize

from sampo.metrics.resources_in_time.base import ResourceOptimizer, init_borders, prepare_answer, is_resources_good
from sampo.scheduler.base import Scheduler
from sampo.schemas.time import Time
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import WorkGraph

MAX_NEWTON_CG_ITERATIONS: int = 48
NEWTON_CG_UNDEFINED_INF: float = 1e+5


@dataclass
class NewtonCGOptimizer(ResourceOptimizer):
    """
    Unsuccessful resource optimizer. Should be reimagined
    """
    scheduler: Scheduler

    gradient_scale: Optional[Union[float, int]] = 10e+5
    worker_weights: Optional[ndarray] = field(default_factory=lambda: defaultdict(lambda: 1))
    worker_factor: Optional[int] = 100
    max_workers: Optional[int] = 500

    _opt_result: ndarray = field(init=False)
    _nfev: int = field(init=False, default=0)  # Number of Function EValuations
    _worker_weights: ndarray = field(init=False)
    _opt_bounds: Tuple[ndarray, ndarray] = field(init=False)
    _margin: int = field(init=False, default=None)

    _can_satisfy_deadline: Callable[[ndarray], bool] = field(init=False, default=None)

    __last: ndarray = field(init=False, default=None)

    def optimize(self, wg: WorkGraph,
                 deadline: Time,
                 worker_weights: Dict[str, Union[int, float]] = None,
                 agents: Optional[WorkerContractorPool] = None,
                 dry_resources: Optional[bool] = False) -> Union[Tuple[Contractor, Time], Tuple[None, None]]:
        self._nfev = 0
        left_counts, right_counts, agent_names = init_borders(wg, self.scheduler, deadline,
                                                              self.worker_factor, self.max_workers, agents)
        self._init_worker_weights(agent_names, worker_weights or self.worker_weights)

        assert left_counts is not None, 'Deadline cannot be satisfied by any set of resources'

        if right_counts is None:
            # TODO: uncomment
            # return prepare_answer(left_counts, agent_names, wg, self.scheduler, False)
            # TODO: remove
            right_counts = left_counts * self.worker_factor

        self._opt_bounds = left_counts, right_counts

        agents = agents or agent_names
        self._can_satisfy_deadline = partial(is_resources_good,
                                             wg=wg, agent_names=agents, scheduler=self.scheduler, deadline=deadline)

        x0, x1 = self._init_x0_x1(left_counts, right_counts)

        opt = scipy.optimize.newton(func=self._target_function,
                                    x0=x0,
                                    x1=x1,
                                    tol=1 / self.gradient_scale,
                                    maxiter=MAX_NEWTON_CG_ITERATIONS)
        self._opt_result = (opt * self.gradient_scale).astype(int)
        print(f'NewtonCG resource optimizer: {self._nfev} fev; {self._opt_result.mean()} mean res')

        return prepare_answer(self._opt_result, agents, wg, self.scheduler, dry_resources)

    def _is_resource_good(self, x: ndarray):
        # TODO: simplify code, when tested
        ib = self._in_bounds(x)
        if ib:
            sd = self._can_satisfy_deadline(agent_counts=x)
            return sd
        return False

    def _init_x0_x1(self, lower_bounds: ndarray, upper_bounds: ndarray):
        # TODO: refactor initializing logics
        result = np.amax((lower_bounds, (lower_bounds + upper_bounds) / 1.75), axis=0)
        ib = self._in_bounds(result)
        rg = self._can_satisfy_deadline(agent_counts=result)
        while not rg and ib:
            result = np.amin((result + 1, upper_bounds), axis=0).astype(int)
            ib = self._in_bounds(result)
            rg = self._can_satisfy_deadline(agent_counts=result)
        return (result + 1) / self.gradient_scale, result / self.gradient_scale

    def _init_worker_weights(self, agent_names: List[str], agent_weights: Dict[str, Union[float, int]]):
        self._worker_weights = np.array([agent_weights[name] for name in agent_names])

    def _in_bounds(self, x: ndarray):
        lower, upper = self._opt_bounds
        return all(x >= lower) and all(x <= upper)

    def _loss(self, x: ndarray):
        lower, _ = self._opt_bounds
        power = 2 if any(x < 0) else 3
        loss = np.abs(x - lower) * self._worker_weights
        if not self._is_resource_good(x):
            return (((loss + 1) * self._worker_weights) ** power).sum()
        return (loss * self._worker_weights).sum()

    def _target_function(self, x: ndarray):
        # TODO: This function doesn't produce proper gradient values and causes fluctuation and non convergence
        scaled_x = (x * self.gradient_scale).astype(int)
        self._nfev += 1

        # result = (scaled_x * self._worker_weights).sum()
        r = self._loss(scaled_x)
        return r




from typing import List, Optional, Dict, Iterable

from sampo.scheduler.base import Scheduler, SchedulerType
from sampo.scheduler.heft.prioritization import prioritization
from sampo.scheduler.heft.time_computaion import calculate_working_time_cascade
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.just_in_time_timeline import JustInTimeTimeline
from sampo.scheduler.utils.multi_contractor import get_worker_borders, run_contractor_search
from sampo.schemas.contractor import Contractor, get_worker_contractor_pool
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.base_opt import dichotomy_int
from sampo.utilities.validation import validate_schedule


class HEFTScheduler(Scheduler):

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.HEFTAddEnd,
                 resource_optimizer: ResourceOptimizer = CoordinateDescentResourceOptimizer(dichotomy_int),
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type, resource_optimizer, work_estimator)

    def schedule_with_cache(self, wg: WorkGraph,
                            contractors: List[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            timeline: Timeline | None = None) \
            -> tuple[Schedule, Timeline]:
        ordered_nodes = prioritization(wg, self.work_estimator)

        schedule, timeline = self.build_scheduler(ordered_nodes, contractors, spec, self.work_estimator, timeline)
        schedule = Schedule.from_scheduled_works(
            schedule,
            wg
        )

        if validate:
            validate_schedule(schedule, wg, contractors)

        return schedule, timeline

    def build_scheduler(self,
                        ordered_nodes: List[GraphNode],
                        contractors: List[Contractor],
                        spec: ScheduleSpec,
                        work_estimator: WorkTimeEstimator = None,
                        timeline: Timeline | None = None) \
            -> tuple[Iterable[ScheduledWork], JustInTimeTimeline]:
        """
        Find optimal number of workers who ensure the nearest finish time.
        Finish time is combination of two dependencies: max finish time, max time of waiting of needed workers
        This is selected by iteration from minimum possible numbers of workers until then the finish time is decreasing
        :param contractors:
        :param work_estimator:
        :param spec: spec for current scheduling
        :param ordered_nodes:
        :param timeline: the previous used timeline can be specified to handle previously scheduled works
        :return:
        """
        worker_pool = get_worker_contractor_pool(contractors)
        # dict for writing parameters of completed_jobs
        node2swork: Dict[GraphNode, ScheduledWork] = {}
        # list for support the queue of workers
        if not isinstance(timeline, JustInTimeTimeline):
            timeline = JustInTimeTimeline(worker_pool)

        for index, node in enumerate(reversed(ordered_nodes)):  # the tasks with the highest rank will be done first
            work_unit = node.work_unit
            work_spec = spec.get_work_spec(work_unit.id)
            if node in node2swork:  # here
                continue

            def run_with_contractor(contractor: Contractor) -> tuple[Time, Time, List[Worker]]:
                min_count_worker_team, max_count_worker_team, workers \
                    = get_worker_borders(worker_pool, contractor, work_unit.worker_reqs)

                if len(workers) != len(work_unit.worker_reqs):
                    return Time(0), Time.inf(), []

                workers = [worker.copy() for worker in workers]

                def get_finish_time(worker_team):
                    return timeline.find_min_start_time(node, worker_team, node2swork) \
                           + calculate_working_time_cascade(node, worker_team, work_estimator)

                # apply worker team spec
                self.optimize_resources_using_spec(work_unit, workers, work_spec,
                                                   lambda optimize_array: self.resource_optimizer.optimize_resources(
                                                       worker_pool, workers,
                                                       optimize_array,
                                                       min_count_worker_team, max_count_worker_team,
                                                       get_finish_time))

                c_st = timeline.find_min_start_time(node, workers, node2swork)
                c_ft = c_st + calculate_working_time_cascade(node, workers, work_estimator)

                return c_st, c_ft, workers

            st, ft, contractor, best_worker_team = run_contractor_search(contractors, run_with_contractor)

            # apply work to scheduling
            timeline.schedule(index, node, node2swork, best_worker_team, contractor,
                              st, work_spec.assigned_time, work_estimator)
            # add using resources in queue for workers
            timeline.update_timeline(index, ft, node, node2swork, best_worker_team)

        # parallelize_local_sequence(ordered_nodes, 0, len(ordered_nodes), work_id2schedule_unit)
        # recalc_schedule(reversed(ordered_nodes), work_id2schedule_unit, worker_pool, work_estimator)

        return node2swork.values(), timeline



from typing import List, Dict, Set, Optional, Iterable

import numpy as np
from toposort import toposort_flatten, toposort

from sampo.scheduler.base import Scheduler
from sampo.scheduler.base import SchedulerType
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.momentum_timeline import MomentumTimeline
from sampo.scheduler.utils.multi_contractor import get_worker_borders, run_contractor_search
from sampo.schemas.contractor import Contractor, get_worker_contractor_pool
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.validation import validate_schedule


class TopologicalScheduler(Scheduler):

    def __init__(self, scheduler_type: SchedulerType = SchedulerType.Topological,
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=AverageReqResourceOptimizer(),
                         work_estimator=work_estimator)

    def schedule_with_cache(self, wg: WorkGraph,
                            contractors: List[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            timeline: Timeline | None = None) \
            -> tuple[Schedule, Timeline]:
        tsorted_nodes: List[GraphNode] = self._topological_sort(wg)

        schedule, timeline = self.build_scheduler(tsorted_nodes, contractors, spec, self.work_estimator, timeline)
        schedule = Schedule.from_scheduled_works(
            schedule,
            wg
        )

        if validate:
            validate_schedule(schedule, wg, contractors)

        return schedule, timeline

    # noinspection PyMethodMayBeStatic
    def _topological_sort(self, wg: WorkGraph) -> List[GraphNode]:
        node2ind = {node: i for i, node in enumerate(wg.nodes)}

        dependents: Dict[int, Set[int]] = {
            node2ind[v]: {node2ind[node] for node in v.parents}
            for v in wg.nodes
        }

        tsorted_nodes_indices: List[int] = toposort_flatten(dependents, sort=True)
        tsorted_nodes = [wg.nodes[i] for i in tsorted_nodes_indices]
        return tsorted_nodes

    # noinspection PyMethodMayBeStatic
    def build_scheduler(self, tasks: List[GraphNode], contractors: List[Contractor],
                        spec: ScheduleSpec,
                        work_estimator: WorkTimeEstimator = None,
                        timeline: Timeline | None = None) \
            -> tuple[Iterable[ScheduledWork], MomentumTimeline]:
        """
        Builds a schedule from a list of tasks where all dependents tasks are guaranteed to be later
        in the sequence than their dependencies
        :param work_estimator:
        :param tasks: list of tasks ordered by some algorithm according to their dependencies and priorities
        :param spec: spec for current scheduling
        :param contractors: pools of workers available for execution
        :param timeline: the previous used timeline can be specified to handle previously scheduled works
        :return: a schedule
        """

        # data structure to hold scheduled tasks
        node2swork: Dict[GraphNode, ScheduledWork] = dict()

        if not isinstance(timeline, MomentumTimeline):
            timeline = MomentumTimeline(tasks, contractors)

        # we can get agents here, because they are always same and not updated
        worker_pool = get_worker_contractor_pool(contractors)

        # We allocate resources for the whole inseparable chain, when we process the first node in it.
        # So, we will store IDs of non-head nodes in such chains to skip them.
        # Note that tasks are already topologically ordered,
        # i.e., the first node in a chain is always processed before its children

        skipped_inseparable_children: Set[str] = set()
        # scheduling all the tasks in a one-by-one manner
        for index, node in enumerate(tasks):
            # skip, if this node was processed as a part of an inseparable chin previously
            if node.id in skipped_inseparable_children:
                continue
            work_unit = node.work_unit
            work_spec = spec.get_work_spec(work_unit.id)

            def run_with_contractor(contractor: Contractor) -> tuple[Time, Time, List[Worker]]:
                min_count_worker_team, max_count_worker_team, workers = \
                    get_worker_borders(worker_pool, contractor, node.work_unit.worker_reqs)

                if len(workers) != len(work_unit.worker_reqs):
                    return Time(0), Time.inf(), []

                worker_team = [worker.copy() for worker in workers]

                # apply worker team spec
                self.optimize_resources_using_spec(work_unit, worker_team, work_spec,
                                                   lambda optimize_array: self.resource_optimizer.optimize_resources(
                                                       worker_pool, worker_team,
                                                       optimize_array,
                                                       min_count_worker_team, max_count_worker_team,
                                                       # dummy
                                                       lambda _: Time(0)))

                c_st, _, exec_times = \
                    timeline.find_min_start_time_with_additional(node, worker_team, node2swork, work_estimator)
                c_ft = c_st
                for node_lag, node_time in exec_times.values():
                    c_ft += node_lag + node_time

                return c_st, c_ft, worker_team

            st, ft, contractor, best_worker_team = run_contractor_search(contractors, run_with_contractor)

            # finish scheduling with time spec
            timeline.schedule(index, node, node2swork, best_worker_team, contractor,
                              st, work_spec.assigned_time, work_estimator)

        return node2swork.values(), timeline


class RandomizedTopologicalScheduler(TopologicalScheduler):
    def __init__(self, work_estimator: Optional[WorkTimeEstimator or None] = None,
                 random_seed: Optional[int] = None):
        super().__init__(work_estimator)
        self._random_state = np.random.RandomState(random_seed)

    def _topological_sort(self, wg: WorkGraph) -> List[GraphNode]:
        def shuffle(nodes: Set[GraphNode]) -> List[GraphNode]:
            nds = list(nodes)
            indices = np.arange(len(nds))
            self._random_state.shuffle(indices)
            return [nds[ind] for ind in indices]

        dependents: Dict[GraphNode, Set[GraphNode]] = {v: v.parents_set for v in wg.nodes}
        tsorted_nodes: List[GraphNode] = [
            node for level in toposort(dependents)
            for node in shuffle(level)
        ]

        # print([node.work_unit.id for node in tsorted_nodes])
        return tsorted_nodes



from random import Random
from typing import Callable

from sampo.generator.config.gen_counts import MIN_GRAPH_COUNTS, ADDITION_CLUSTER_PROBABILITY, GRAPH_COUNTS, \
    MAX_BOREHOLES_PER_BLOCK, BRANCHING_PROBABILITY
from sampo.generator.pipeline.cluster import get_start_stage, get_cluster_works, get_finish_stage, add_addition_work
from sampo.generator.types import SyntheticGraphType, StageType
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.utils import count_node_ancestors


def get_small_graph(cluster_name: str | None = 'C1', rand: Random | None = None) -> WorkGraph:
    """
    Creates a small graph of works consisting of 30-50 vertices;
    :param cluster_name: str - the first cluster name
    :param rand: Optional[Random] - generator of numbers with a given seed or None
    :return:
       work_graph: WorkGraph - work graph where count of vertex between 30 and 50
    """

    s = get_start_stage()
    pipe_nodes_count = MIN_GRAPH_COUNTS['pipe_nodes'].rand_int(rand)
    pipe_net_count = MIN_GRAPH_COUNTS['pipe_net'].rand_int(rand)
    light_masts_count = MIN_GRAPH_COUNTS['light_masts'].rand_int(rand)
    borehole_count = MIN_GRAPH_COUNTS['borehole'].rand_int(rand)

    c1, _ = get_cluster_works(root_node=s, cluster_name=cluster_name, pipe_nodes_count=pipe_nodes_count,
                              pipe_net_count=pipe_net_count, light_masts_count=light_masts_count,
                              borehole_counts=[borehole_count], rand=rand)
    f = get_finish_stage([c1])
    graph = WorkGraph(s, f)
    return graph


def get_cluster_graph(root_node: GraphNode, cluster_name: str, pipe_nodes_count: int | None = None,
                      pipe_net_count: int | None = None, light_masts_count: int | None = None,
                      borehole_counts: list[int] | None = None, add_addition_cluster: bool | None = False,
                      addition_cluster_probability: float | None = ADDITION_CLUSTER_PROBABILITY,
                      rand: Random | None = None) -> (list[GraphNode], dict[str, GraphNode]):
    pipe_nodes_count = pipe_nodes_count or GRAPH_COUNTS['pipe_nodes'].rand_int(rand)
    pipe_net_count = pipe_net_count or GRAPH_COUNTS['pipe_net'].rand_int(rand)
    light_masts_count = light_masts_count or GRAPH_COUNTS['light_masts'].rand_int(rand)
    if borehole_counts is None:
        counts = GRAPH_COUNTS['borehole'].rand_int(rand)
        if counts <= MAX_BOREHOLES_PER_BLOCK:
            borehole_counts = [counts]
        else:
            borehole_counts = [MAX_BOREHOLES_PER_BLOCK, counts - MAX_BOREHOLES_PER_BLOCK]

    c_master, roads = get_cluster_works(root_node=root_node, cluster_name=cluster_name,
                                        pipe_nodes_count=pipe_nodes_count, pipe_net_count=pipe_net_count,
                                        light_masts_count=light_masts_count, borehole_counts=borehole_counts, rand=rand)

    checkpoints = [c_master]
    if add_addition_cluster or add_addition_work(addition_cluster_probability, rand):
        c_slave, _ = get_cluster_works(root_node=root_node, cluster_name=cluster_name,
                                       pipe_nodes_count=pipe_nodes_count, pipe_net_count=pipe_net_count,
                                       light_masts_count=light_masts_count, borehole_counts=borehole_counts,
                                       roads=roads, rand=rand)
        checkpoints.append(c_slave)
    return checkpoints, roads


def get_graph(mode: SyntheticGraphType | None = SyntheticGraphType.General,
              cluster_name_prefix: str | None = 'C',
              cluster_counts: int | None = 0,
              branching_probability: float | None = BRANCHING_PROBABILITY,
              addition_cluster_probability: float | None = ADDITION_CLUSTER_PROBABILITY,
              bottom_border: int | None = 0,
              top_border: int | None = 0,
              rand: Random | None = None) -> WorkGraph:
    """
    Invokes a graph of the given type if at least one positive value of
        cluster_counts, addition_cluster_probability, bottom_border is given;
    :param mode: str - 'general' or 'sequence' or 'parallel - the type of the returned graph
    :param cluster_name_prefix: str -  cluster name prefix, if the prefix is 'C',
        the clusters will be called 'C1', 'C2' etc.
    :param cluster_counts: Optional[int] - Number of clusters for the graph
    :param branching_probability: Optional[float] - The probability that the node will not be connected to the last
        cluster, but to any other cluster for the general mode
    :param addition_cluster_probability: Optional[float] - probability of a slave (example C3_1) cluster
        to the main cluster
    :param bottom_border: Optional[int] - bottom border for number of works for the graph
    :param top_border: Optional[int] - top border for number of works for the graph
    :param rand: Optional[Random] - generator of numbers with a given seed or None
    :return:
        work_graph: WorkGraph - the desired work graph
    """
    assert cluster_counts + bottom_border + top_border > 0, 'At least one border param should be specified'
    assert cluster_counts >= 0 and branching_probability >= 0 and top_border >= 0, 'Params should not be negative'

    rand = rand or Random()
    get_root_stage: Callable[[list[StageType], float, Random], GraphNode] = graph_mode_to_callable(mode)

    if bottom_border > 0:
        top_border = 0
    masters_clusters_ind = 1
    works_generated = 0
    s = get_start_stage()
    stages: list[StageType] = [(s, {'min': s})]

    while True:
        if cluster_counts > 0 and (cluster_counts == len(stages) - 1):
            addition_cluster_probability = 0

        root_stage = get_root_stage(stages, branching_probability, rand)
        checkpoints, r = get_cluster_graph(root_stage, f'{cluster_name_prefix}{masters_clusters_ind}',
                                           addition_cluster_probability=addition_cluster_probability, rand=rand)
        tmp_finish = get_finish_stage(checkpoints)
        count_works = count_node_ancestors(tmp_finish, root_stage)

        if 0 < top_border < (count_works + works_generated):
            break

        stages += [(c, r) for c in checkpoints]
        masters_clusters_ind += 1
        works_generated += count_works

        if 0 < bottom_border <= works_generated or 0 < cluster_counts <= (len(stages) - 1):
            break

    if len(stages) == 1:
        return get_small_graph(cluster_name=f'{cluster_name_prefix}1')
    f = get_finish_stage([c for c, _ in stages[1:]])
    graph = WorkGraph(s, f)
    return graph


def graph_mode_to_callable(mode: SyntheticGraphType) -> \
        Callable[[list[tuple[GraphNode, dict[str, GraphNode]]], float, Random], GraphNode]:
    if mode is SyntheticGraphType.General:
        return general_graph_mode
    if mode is SyntheticGraphType.Parallel:
        return parallel_graph_mode_get_root
    return sequence_graph_mode_get_root


def parallel_graph_mode_get_root(stages: list[tuple[GraphNode, dict[str, GraphNode]]],
                                 branching_probability: float, rand: Random) -> GraphNode:
    return stages[0][0]


def sequence_graph_mode_get_root(stages: list[tuple[GraphNode, dict[str, GraphNode]]],
                                 branching_probability: float, rand: Random) -> GraphNode:
    return stages[-1][0]


def general_graph_mode(stages: list[tuple[GraphNode, dict[str, GraphNode]]],
                       branching_probability: float, rand: Random) -> GraphNode:
    is_branching = rand.random() <= branching_probability
    ind = len(stages) - 1
    if is_branching and len(stages) > 2:
        ind = rand.randint(1, len(stages) - 2)
    return stages[ind][1]['min']
